const {yellow, green, red} = require('chalk');
const path = require('path');
const createRollupConfig = require('../config/rollup.config.default');
const rollup = require('rollup');
const express = require('express');
const mashupRollupPlugin = require('../plugins/mashupRollupPlugin');
const cors = require('cors');

const runRollupDevServer = async function(
  dir = process.cwd(),
  {minify = false, dependencyReplace, outputDir, port = 9090}
) {
  outputDir = outputDir || path.join(dir, 'dist');
  const basePath = `http://localhost:${port}`;
  const config = await createRollupConfig({
    dir,
    minify,
    dependencyReplace,
    outputDir,
    basePath
  });
  const {
    settings: {mashupModuleName}
  } = config.plugins.find(
    plugin => plugin.name === mashupRollupPlugin.pluginName
  );
  const server = express();
  server.use(cors());
  server.use(express.static(outputDir));
  server.set('view engine', 'pug');
  server.set('views', path.join(__dirname, '../config'));
  server.use((request, response, next) => {
    if (request.url === '/') {
      const bundleAbsoluteLink = config.input[0]
      const bundleExtension = path.extname(bundleAbsoluteLink)

      response.render('index', {
        name: 'create-mashup-app dev server',
        moduleName: mashupModuleName,
        bundleLink: path.basename(bundleAbsoluteLink, bundleExtension) + '.js'
      });
    } else {
      next();
    }
  });
  let serverStarting = false;
  let firstBuildPassed = false;

  const watcher = rollup.watch(config);
  watcher.on('event', event => {
    switch (event.code) {
      case 'START': {
        if (!serverStarting) {
          serverStarting = true;
          server.listen(port);
        }
        break;
      }
      case 'BUNDLE_START': {
        console.log(
          `Building ${yellow(event.input[0])} -> ${green(event.output[0])}...`
        );
        break;
      }
      case 'END': {
        if (!firstBuildPassed) {
          firstBuildPassed = true;
          console.log(`Server listening at ${yellow(basePath)}`);
        }
        break;
      }
      case 'ERROR': {
        console.log(red(event.error.message));
        break;
      }
      case 'FATAL': {
        throw event.error;
      }
    }
  });
};

module.exports = async (...args) => {
  try {
    await runRollupDevServer(...args);
  } catch (err) {
    console.error(red(err));
    console.error(err.stack);
    process.exit(-1);
  }
};
