const yeoman = require('yeoman-environment');
const Generator = require('yeoman-generator');
const {red} = require('chalk');
const path = require('path');
const {version} = require('../../package.json');

class MashupAppGenerator extends Generator {
  paths() {
    this.sourceRoot(path.join(__dirname, '../templates'));
  }

  async prompting() {
    const {packageName} = await this.prompt([
      {
        type: 'input',
        name: 'packageName',
        message: 'Your mashup package name:',
        default: this.appname.replace(/\s+/g, '-')
      }
    ]);

    const {moduleName, displayName, port} = await this.prompt([
      {
        type: 'input',
        name: 'moduleName',
        message: 'Your mashup module name:',
        default: `tp3/integrations/${packageName}`
      },
      {
        type: 'input',
        name: 'displayName',
        message: 'Your mashup display (user-friendly) name:',
        default: packageName.split('-').join(' ')
      },
      {
        type: 'input',
        name: 'port',
        message: 'Default port for dev server:',
        default: '3001'
      }
    ]);

    const {template} = await this.prompt({
      type: 'list',
      name: 'template',
      message: 'Template',
      choices: [
        {
          name:
            'Minimal (bare minimum, no additional overhead, install as you go)',
          value: 'base'
        },
        {
          name:
            'React (included: simple component, test config & simple working test)',
          value: 'react'
        },
        {
          name:
            'Minimal typescript (same as minimal with typescript and basic tsconfig.json)',
          value: 'minimal-typescript'
        }
      ]
    });

    this.params = {
      packageName,
      moduleName,
      displayName,
      createMashupAppVersion: version,
      template,
      port
    };
  }

  writing() {
    this.fs.copyTpl(
      this.templatePath('base/**/*'),
      this.destinationPath(),
      this.params,
      {},
      {globOptions: {dot: true}}
    );

    if (this.params.template !== 'base') {
      this.fs.copyTpl(
        this.templatePath(`${this.params.template}/**/*`),
        this.destinationPath(),
        this.params,
        {},
        {globOptions: {dot: true}}
      );
    }

    if (this.params.template === 'minimal-typescript') {
      this.fs.move(
        this.destinationPath('src/index.js'),
        this.destinationPath('src/index.ts')
      );
    }
  }

  install() {
    this.installDependencies({npm: true, bower: false, yarn: false});

    if (this.params.template === 'react') {
      this.npmInstall(
        [
          'eslint',
          'babel-eslint',
          'stylelint',
          'stylelint-config-standard',
          'jest@^21.2',
          'enzyme@^3.2',
          'enzyme-adapter-react-15@^1.0.5',
          'identity-obj-proxy',
          'jest-environment-jsdom-debug@^1.0.1',
          'react-test-renderer@15',
          'eslint-plugin-babel',
          'eslint-plugin-react'
        ],
        {'save-dev': true}
      );
    }
  }

  constructor(args, options) {
    super(args, options);

    this.sourceRoot(path.join(__dirname, 'templates'));
  }
}

async function generateApp() {
  const yoEnv = yeoman.createEnv();

  yoEnv.registerStub(MashupAppGenerator, 'create-mashup-app-generator');

  return new Promise(resolve => {
    yoEnv.run('create-mashup-app-generator', () => {
      console.log('Generator executed');
      resolve(0);
    });
  });
}

module.exports = async (...args) => {
  try {
    await generateApp(...args);
    process.exit(0);
  } catch (err) {
    console.error(red(err));
    console.error(err.stack);
    process.exit(-1);
  }
};
