const createRollupConfig = require('../../src/config/rollup.config.default');
const path = require('path');
const rollup = require('rollup');
const jsdom = require('jsdom');
const validate = require('sourcemap-validator');

jest.setTimeout(9000000);
describe('the bundling process', () => {
  afterEach(() => {
    const styleTags = Array.from(document.getElementsByTagName('style'));
    styleTags.forEach(tag => tag.parentNode.removeChild(tag));
  });

  it('should obey minification flag', async () =>
    Promise.all(
      [true, false].map(async minify => {
        const config = await createRollupConfig({
          dir: path.join(__dirname, 'minify'),
          minify
        });

        const bundle = await rollup.rollup(config);
        const {output} = await bundle.generate(config.output);
        const {code} = output['mashup.js'];
        const mangledVariableName = 'thisVariableNameShouldBeMangledByMinifier';
        if (minify) {
          expect(code.indexOf(mangledVariableName)).toBeLessThan(0);
        } else {
          expect(code.indexOf(mangledVariableName)).toBeGreaterThan(0);
        }
      })
    ));
  it('should remove all comments from the bundle, even if minify is false', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'require-in-comments'),
      minify: false
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];
    expect(code.indexOf('require("in-comment")')).toBeLessThan(0);
    expect(code.indexOf('require("in-string")')).toBeGreaterThan(0);
  });
  it('should produce a bundle that calls into mashups API with addMashup or addModule and correctly handles dependencies', async () =>
    Promise.all(
      ['mashup'].map(async target => {
        const config = await createRollupConfig({
          dir: path.join(__dirname, 'dependency'),
          target,
          minify: true
        });

        const bundle = await rollup.rollup(config);
        const {output} = await bundle.generate(config.output);
        const {code} = output['mashup.js'];

        if (window.tau) {
          delete window.tau;
        }
        const notRemappedMock = jest.fn();
        const remappedMock = jest.fn();
        let moduleResult;

        // fake global will be called in actual mashup code to verify that it executed
        window.tau = {mashups: {}, fakeGlobal: jest.fn()};
        window.tau.mashups.addDependency = jest.fn(() => window.tau.mashups);
        window.tau.mashups.addMashup = jest.fn(initializer =>
          initializer(notRemappedMock, remappedMock)
        );
        window.tau.mashups.addModule = jest.fn((moduleName, initializer) => {
          moduleResult = initializer(notRemappedMock, remappedMock);
        });

        // jsdom here, so we can eval generated mashup code straight away
        eval(code);

        expect(window.tau.mashups.addDependency).toHaveBeenCalledWith(
          'not-remapped-dependency'
        );
        expect(window.tau.mashups.addDependency).toHaveBeenCalledWith(
          'remapped-dependency'
        );

        if (target === 'mashup') {
          expect(window.tau.mashups.addMashup).toHaveBeenCalled();
          // mashups should be executed right away
        } else {
          expect(window.tau.addModule).toHaveBeenCalled();

          // modules should not execute their code until their initialize is called
          expect(notRemappedMock).not.toHaveBeenCalled();
          expect(remappedMock).not.toHaveBeenCalled();
          expect(window.tau.fakeGlobal).not.toHaveBeenCalled();

          expect(typeof moduleResult.initialize).toBe('function');

          moduleResult.initialize();
        }

        expect(notRemappedMock).toHaveBeenCalled();
        expect(remappedMock).toHaveBeenCalled();
        expect(window.tau.fakeGlobal).toHaveBeenCalled();
      })
    ));

  it('should add json files to bundle', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'rollup-json'),
      target: 'mashup',
      minify: true
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];
    if (window.tau) {
      delete window.tau;
    }

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}, dumpStyles: jest.fn()};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    eval(code);

    expect(window.data).toEqual({
      foo: 'bar'
    });
  });

  it('should produce a bundle that uses CSS modules correctly and appends all used classes as style tag', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'css-modules'),
      target: 'mashup',
      minify: true
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}, dumpStyles: jest.fn()};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    // jsdom here, so we can eval generated mashup code straight away
    eval(code);

    expect(window.tau.dumpStyles).toHaveBeenCalledWith({
      ['test-class']: 'test-class__css-modules__1-0-0',
      testClass: 'test-class__css-modules__1-0-0',
      foo: 'foo__css-modules__1-0-0',
      bar: 'bar__css-modules__1-0-0',
      imported: 'imported__css-modules__1-0-0',
      nested: 'nested__css-modules__1-0-0'
    });

    const styleTag = document.getElementsByTagName('style')[0];
    expect(
      styleTag.innerHTML.indexOf('.test-class__css-modules__1-0-0')
    ).toBeGreaterThanOrEqual(0);
    expect(styleTag.innerHTML.indexOf('.global')).toBeGreaterThanOrEqual(0);
    expect(
      styleTag.innerHTML.indexOf('.global__css-modules__1-0-0')
    ).toBeLessThan(0);
  });

  it('should throw when dynamic imports are used and basePath is not specified', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'code-splitting'),
      target: 'mashup',
      minify: true
    });

    const bundle = await rollup.rollup(config);
    await expect(bundle.generate(config.output)).rejects.toThrow(
      /basePath not specified/
    );
  });

  it('should perform code-splitting and load chunks when they are required', async () => {
    const chunksBasePath = 'http://fake-url';
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'code-splitting'),
      target: 'mashup',
      basePath: chunksBasePath,
      minify: true
    });

    const {mashupChunksGlobalName} = config.plugins.find(
      plugin => plugin.name === 'mashup-entry-plugin'
    ).settings;

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const chunkNames = Object.keys(output);
    // + mashup.js (entry point) includes 2 common chunks - commonChunk.js
    // used by chunk1.js and chunk2.js and regenerator runtime used by
    // entry and arbitaryFolder/chunk
    // + chunk1.js
    // + chunk2.js
    // + chunk.js
    // + arbitaryFolder/chunk.js (name clashing is intentional)
    // + manifest.cfg (because target: 'mashup')
    expect(chunkNames).toHaveLength(6);

    // we can't use Jest's built-in jsdom because we need to override script loading options
    const chunkRequests = [];
    const dom = new jsdom.JSDOM('', {
      runScripts: 'dangerously',
      resources: 'usable',
      resources: new class extends jsdom.ResourceLoader {
        fetch(url) {
          if (url.startsWith(chunksBasePath)) {
            const chunkName = url.substring(chunksBasePath.length + 1);
            chunkRequests.push(chunkName);
            return Promise.resolve(Buffer.from(output[chunkName].code));
          }
        }
      }()
    });

    const {window} = dom;

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}, dumpStyles: jest.fn()};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();
    const initChunkCode = output['mashup.js'].code;
    // we expect our entry point to also contain commons chunk - in this case, regenerator runtime
    expect(initChunkCode.indexOf('common chunk')).toBeGreaterThan(0);

    await new Promise(resolve => {
      window.tau.initChunksLoaded = resolve;
      window.eval(initChunkCode);
      // by the time entry chunk is executed on the page, we expect to have 3
      // chunks available: mashup.js (entry), common chunk with regenerator and
      // common chunk from commonChunk.js
      expect(Object.keys(window[mashupChunksGlobalName])).toHaveLength(3);
    });
    // when the above promise resolves, we can check loaded chunks
    // we expect all chunks except the last one to be present
    expect(Object.keys(window[mashupChunksGlobalName])).toHaveLength(6);
    await new Promise(resolve => {
      window.tau.lastChunkLoaded = resolve;
      window.tau.loadLastChunk();
    });

    // the last chunk is loaded only after arbitary lastChunkLoaded function call
    // there are 7 chunks total on the page - entry, 2 common and 4 from dynamic imports
    expect(Object.keys(window[mashupChunksGlobalName])).toHaveLength(7);
    // we expect dynamic imports to be reused once they are loaded, i.e.
    // there should be no more than 4 requests for dynamic chunks
    expect(chunkRequests).toHaveLength(4);
  });

  it('should support typescript and tsx out of the box', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'typescript'),
      target: 'mashup',
      minify: true
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }
    const fakeComponentConstructor = jest.fn();
    const fakeReact = {
      createElement: jest.fn((type, props, children) => ({
        type,
        props,
        children
      })),
      Component: function(...args) {
        fakeComponentConstructor(...args);
      }
    };

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer(fakeReact);

    eval(code);
    expect(fakeReact.createElement).toHaveBeenCalledWith('div', null, {
      type: window.tau.smallComponent,
      children: undefined,
      props: {id: 5}
    });
    expect(fakeComponentConstructor).toHaveBeenCalledWith({name: 'foo'});
    expect(window.tau.reactElement.type).toBe('div');
    expect(Object.getPrototypeOf(window.tau.reactElement.children.type)).toBe(
      fakeReact.Component
    );
  });

  it('should correctly resolve dependencies from node_modules', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'node-resolve'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    window.tau.assertFoo = jest.fn();

    eval(code);

    expect(window.tau.assertFoo).toHaveBeenCalledTimes(2);
    expect(window.tau.assertFoo).toHaveBeenNthCalledWith(1, 'foo');
    expect(window.tau.assertFoo).toHaveBeenNthCalledWith(2, 'foo');
  });

  it('should correctly resolve dependencies from node_modules', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'node-resolve'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    window.tau.assertFoo = jest.fn();

    eval(code);

    expect(window.tau.assertFoo).toHaveBeenCalledTimes(2);
    expect(window.tau.assertFoo).toHaveBeenNthCalledWith(1, 'foo');
    expect(window.tau.assertFoo).toHaveBeenNthCalledWith(2, 'foo');
  });

  it('should replace process.env.NODE_ENV with correct string based on minify', async () =>
    Promise.all(
      [true, false].map(async minify => {
        const config = await createRollupConfig({
          dir: path.join(__dirname, 'process-env'),
          target: 'mashup',
          minify
        });

        const bundle = await rollup.rollup(config);
        const {output} = await bundle.generate(config.output);
        const {code} = output['mashup.js'];

        // check that DCE works and condition is removed completely
        const hasAssertProductionCall =
          code.indexOf('window.tau.assertProduction()') > 0;
        const hasAssertDevelopmentCall =
          code.indexOf('window.tau.assertDevelopment()') > 0;
        expect(hasAssertProductionCall).toBe(minify);
        expect(hasAssertDevelopmentCall).toBe(!minify);

        if (window.tau) {
          delete window.tau;
        }

        // fake global will be called in actual mashup code to verify that it executed
        window.tau = {mashups: {}};
        window.tau.mashups.addDependency = () => window.tau.mashups;
        window.tau.mashups.addMashup = initializer => initializer();

        window.tau.assertProduction = jest.fn();
        window.tau.assertDevelopment = jest.fn();

        eval(code);

        if (minify) {
          expect(window.tau.assertProduction).toHaveBeenCalled();
          expect(window.tau.assertDevelopment).not.toHaveBeenCalled();
        } else {
          expect(window.tau.assertDevelopment).toHaveBeenCalled();
          expect(window.tau.assertProduction).not.toHaveBeenCalled();
        }
      })
    ));

  it('should fail build for typescript-specific compilation errors', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'typescript-errors')
    });

    await expect(rollup.rollup(config)).rejects.toThrow(
      /Argument of type '10' is not assignable to parameter of type 'string'/
    );
  });

  it('should allow resolution with absolute path relative to source dir', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'relative-resolution'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);

    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    eval(code);

    const {imports} = window.tau;

    expect(typeof imports.import1.foo).toBe('function');
    expect(typeof imports.import2.bar).toBe('function');
    expect(imports.style1.foo).toBe('foo__relative-resolution__1-0-0');
    expect(imports.style2.bar).toBe('bar__relative-resolution__1-0-0');
  });

  it('should transform typescript output with babel', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'babel-after-typescript'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);

    const {code} = output['mashup.js'];

    expect(code.indexOf('const')).toBeLessThan(0);
    expect(code.indexOf('class Foo')).toBeLessThan(0);
  });

  it('should transform typescript output with babel', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'postcss-vars'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);

    const {code} = output['mashup.js'];

    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    eval(code);

    const {cssModule} = window.tau;

    const styleTag = document.getElementsByTagName('style')[0];

    expect(cssModule).toMatchObject({
      foo: 'foo__postcss-vars__1-0-0'
    });

    expect(styleTag.innerHTML.indexOf('width: 5px')).toBeGreaterThan(0);
    expect(styleTag.innerHTML.indexOf('color: red')).toBeGreaterThan(0);
  });

  it('should replace Object.assign and object spread with runtime call', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'object-assign-transform'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);

    const {code} = output['mashup.js'];

    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();
    delete Object.assign;

    eval(code);

    const {assigned, spread} = window.tau;

    expect(assigned).toMatchObject({
      foo: 'foo',
      bar: 'baz'
    });

    expect(spread).toMatchObject({
      foo: 'foo',
      bar: 'baz'
    });
  });

  it('should return valid mapping object', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'minify'),
      minify: false
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);

    const {code, map} = output['mashup.js'];
    expect(() => {
      validate(code, map);
    }).not.toThrowError();
  });

  it('should allow semver in package.json', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'semver')
    });

    const bundle = await rollup.rollup(config);
    await expect(bundle.generate(config.output)).resolves.toBeTruthy();
  });
});
